-- Time-stamp: <Mon Dec 17 2007 22:55:30 Stardate: [-29]8674.56 hwloidl>
--
-- Interpreter for the Intermediate Language of the MiniJava compiler
---------------------------------------------------------------------------

module Interp where

import AbsSyn
import PPIntm
import Lex
import Parse
import Data.Map as Map
import Control.Monad
import Text.Regex
import Maybe
import List
import Control.Monad.State
import Control.Monad.Reader

import Debug.Trace(trace)

type RegMap = Map Register Val
type Frame = Integer
type Offset = Integer
type Memory = Map Address Val
type Address = Integer
type Trace = String
data Val = I !Integer |
           A !Address |
           L !Label   |
           UNDEF
     deriving (Read,Eq)
type Msgs = [String]

data TraceOpts = CallTrace 
               | ParamTrace 
               | JumpTrace 
               | MoveTrace 
               | AllocTrace 
               | RegMaps 
               | MemMaps 
               | LabMaps 
               | StdCallConv 
               | Warnings 
               | Debug 
               | ShowAST
               | Verbose 
               | None
 deriving (Read,Show,Eq)
type Settings = ([TraceOpts], Int {-level-})

type LabTab = Map Label [Stm]
type FunTab = Map Label [Stm]
type FSTab = Map Label (Maybe Integer) -- table of framesizes
type FunParams = Map Label [Access]
type SpecialRegs = (String,String,String,String,Integer,Integer)

type InterpState = (RegMap,Memory,LabTab,FunTab,FSTab,FunParams,Msgs,Settings,Cost,SpecialRegs)

-- evaluation monad
data Intp a = T (InterpState -> (InterpState, a))

runIntp :: InterpState -> Intp a -> a
runIntp s (T f) = snd (f s)

instance Monad Intp where
 return x = T (\ s -> (s, x))
 (T t) >>= f =  T (\ s -> let { (s1,x1) = t s ; T f' = f x1 } in f' s1)

-- to avoid output like "I (-999)"
instance Show Val where
  show (I i) = "I " ++ show i
  show (A i) = "A " ++ show i
  show (L i) = "L " ++ show i
  show UNDEF = "UNDEF"
 
instance Num Val where
 (A x) + (A y) = A (x+y)
 (A x) + (I y) = A (x+y)
 (I x) + (A y) = A (x+y)
 (I x) + (I y) = I (x+y)
 (A x) * (A y) = A (x*y)
 (A x) * (I y) = A (x*y)
 (I x) * (A y) = A (x*y)
 (I x) * (I y) = I (x*y)
 UNDEF * _ = error ("'*' applied to undefined value")
 _ * UNDEF = error ("'*' applied to undefined value")
 abs (I x) = I (abs x)
 abs (A _) = error "Cannot apply abs to an address"
 abs UNDEF =  error ("'abs' applied to undefined value")
 signum (I x) = I (signum x)
 signum (A _) = error "Cannot apply signum to an address"
 signum UNDEF =  error ("'signum' applied to undefined value")
 fromInteger n = I n

instance Ord Val where
 (A x) <= (A y) = x<=y
 (I x) <= (I y) = x<=y
 UNDEF <= _ =  error ("'<=' applied to undefined value")
 _ <= UNDEF =  error ("'<=' applied to undefined value")
 _ <= _ = error "Cannot compare address with integer"
 
emptyEnv :: RegMap
emptyEnv = empty

emptyTrace :: Msgs
emptyTrace = []

defaultSettings :: Settings
defaultSettings = ([CallTrace,JumpTrace,MoveTrace,AllocTrace,ParamTrace,LabMaps], 0)

undefinedRegs = ("","","","",0,0)

regsWithDefSizes = ("","","","",defSpSz,defHpSz)

defaultRegs = ("tfp","tsp","thp","trp",0,0)
--            (fp   ,sp     ,hp     ,rp  ,ssz',hsz')

-- left-biased
combineRegs (fp',sp',hp',rp',ssz',hsz') (fp,sp,hp,rp,ssz,hsz) =
 (if fp'=="" then fp else fp',
  if sp'=="" then sp else sp',
  if hp'=="" then hp else hp',
  if rp'=="" then rp else rp',
  if ssz'==0 then ssz else ssz',
  if hsz'==0 then hsz else hsz')

emptyState :: [TraceOpts] -> SpecialRegs -> InterpState
emptyState topts regs = (Map.empty, Map.empty, 
              	    Map.empty, Map.empty, Map.empty, Map.empty, 
              	    emptyTrace, (topts, 0), 0, regs)

combineStates (rm, mm, lT, fT, fsT, pT, ts, ss@(trc,lvl), c, r) (rm', mm', lT', fT', fsT', pT', ts', ss'@(trc',lvl'), c', _) =
  (Map.union rm rm', Map.union mm mm', Map.union lT lT', Map.union fT fT', Map.union fsT fsT',  Map.union pT pT', ts++ts', (trc++trc', max lvl lvl'), c+c', r)

-- type State = (RegMap,Memory,LabTab,FunTab,FunParams,Msgs,Settings)
initState :: [TraceOpts] -> SpecialRegs -> [AbsSyn.Fct] -> InterpState
initState topts regs = (foldl combineStates (emptyState topts regs)) .
            (Prelude.map ( \ (Fct name ps fs ss) -> (Map.empty, Map.empty, -- empty regs and heap
	                      runLabTabFiller (fillTab ss),
                              Map.union Map.empty (Map.singleton name ss), -- body
                              Map.union Map.empty (Map.singleton name fs),
                              Map.union Map.empty (Map.singleton name (Prelude.map param2access ps)), -- params
                              emptyTrace, ([], 0), 0, ("","","","")) ))

-- dis-abled trace
trc s e = e

-- retReg :: String 
-- retReg = "t2"

-- retRegA :: Access
-- retRegA = InReg retReg

-- names of callee save registers (UNUSED)
calleeSaves :: [String]
calleeSaves = ["t40","t41","t42","t43"]

-- thisReg :: Intp String 
-- thisReg =  T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,(_,_,_,tp,_,_)) -> (st,tp)

-- thisRegA :: Access
-- thisRegA = InReg thisReg

-- frame pointer
fpReg :: Intp String 
fpReg = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,(fp,_,_,_,_,_)) -> (st,fp)

-- stack pointer
spReg :: Intp String 
spReg = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,(_,sp,_,_,_,_)) -> (st,sp)

-- heap pointer
hpReg :: Intp String 
hpReg = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,(_,_,hp,_,_,_)) -> (st,hp)

-- register for return values of functions
retReg :: Intp String 
retReg =  T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,(_,_,_,rp,_,_)) -> (st,rp)

spSize :: Intp Integer
spSize = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,(_,_,_,_,sz,_)) -> (st,sz)

hpSize :: Intp Integer
hpSize = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,(_,_,_,_,_,sz)) -> (st,sz)


-- dummy vals; mainly to catch errors of uninit vars
dummyInt = UNDEF -- (I 9996)
dummyAddr = UNDEF -- (A 0)

dummyRegVal = UNDEF -- (I 8881)
dummyMemVal = UNDEF -- (I 8882)

spStart = (A 8000)
hpStart = (A 8004)
spEnd = (A 1024)
hpEnd = (A 64*65535)
defSpSz = 7900
defHpSz = 64*65535-8004

globalFrameSizeHACK = 20
---

showMap :: (Show k, Show v) => Map k v -> String
showMap = unlines . (Prelude.map (\ (k, v) -> show k ++ "\t= " ++ (show v))) . toList 

showRegMap :: Intp ()
showRegMap = getR >>= \ rm -> trace ("Register Mapping: " ++ showMap rm) (return ())
--showRegMap = getR >>= \ rm -> msgs ["Register Mapping: ", (showTree rm)]

showMemMap :: Intp ()
showMemMap = getM >>= \ rm -> trace ("Memory Mapping: " ++ (showMap rm)) (return ())
--showMemMap = getM >>= \ rm -> msgs ["Memory Mapping: ", (showTree rm)]

showLabTab :: Intp ()
showLabTab = labTab >>= \ lT -> trace ("labTab: \n" ++ (unlines (Prelude.map (\ (k, a) -> k ++ " |->\n" ++ render (pp a)) (toList lT)))) (return ())
--showLabTab = labTab >>= \ lT -> msgs ["labTab: ", (showTreeWith (\ k a -> k ++ ":=" ++ render (pp a)) True False lT)]

showFunTab :: Intp ()
showFunTab = funTab >>= \ fT -> trace ("funTab: \n" ++ (unlines (Prelude.map (\ (k, a) -> k ++ " |->\n" ++ render (pp a)) (toList fT)))) (return ())
--showFunTab = funTab >>= \ fT -> msgs ["funTab: ", ((showTreeWith (\ k a -> k ++ ":=" ++ render (pp a)) False False fT))]

showParamTab :: Intp ()
showParamTab = paramTab >>= \ fT -> trace ("paramTab: " ++ ((showMap (mapWithKey (\ k a -> k ++ ":=" ++ (show a)) fT)))) (return ())

-- ---------------------------------------------------------------------------
-- wrapper fcts to use the interpreter from within ghci

readPrg :: String -> IO ([TraceOpts], [(String, Either String Integer)], SpecialRegs, [AbsSyn.Fct], AbsSyn.Fct, [Integer])
readPrg prgFile = 
 do
  let topts = [Verbose]
  
  input <- readFile prgFile
  -- putStrLn $ "Passing these args to main fct: " ++ (show args)
  -- input <- getContents
  -- print (alexScanTokens input)
  let (defs,ast) = parse (alexScanTokens input)
  when (Interp.ShowAST `elem` topts) $
   putStrLn ("AST:\n" ++ (render $ pp ast))
  --let res = Interp.runIntp (Interp.initState ast) (Interp.eval 
  let rx = mkRegex "^.*\\$"
  -- let xs = map (\ (AbsSyn.Fct name _ _) -> (subRegex rx name "")) ast
  -- putStrLn (unlines xs)
  let main_ast = case (find (\ (AbsSyn.Fct name _ _ _) -> (subRegex rx name "") == "main") ast) of 
   		   Nothing -> error "PANIC: no main function found" 
   		   (Just main_ast) -> main_ast 

  return (topts, defs, defaultRegs, ast, main_ast, [])

execCall :: ([TraceOpts], [(String, Either String Integer)], SpecialRegs, [AbsSyn.Fct], AbsSyn.Fct, [Integer]) -> IO ()
execCall (topts, defs, regs, ast, main_ast, args) =
 do
  let (msgs,c,m,res) = execCall_with_env topts defs regs ast main_ast args
  putStrLn ("Messages:\n" ++ unlines msgs)
  -- putStrLn ("Result:\n" ++ (show res))
  putStrLn ("Execution time: " ++ (show c) ++ " cycles;\t Heap size: " ++ (show m) ++ " cells")

-- =============================================================================
-- top-level function for interpreter

execCall_with_env :: [TraceOpts] -> [(String, Either String Integer)] -> SpecialRegs -> [AbsSyn.Fct] -> AbsSyn.Fct -> [Integer] -> ([String],Cost,Int,Val)
execCall_with_env topts defs cmdline_regs  prg (Fct name ps _ ss) es = 
  let
       codeOk = evalState (mapM checkCode prg >>= \ oks -> return (foldl bonzo Nothing oks))
                          (CCST { labTabCC=Map.empty, 
                                  funTabCC=(Map.fromList (Prelude.map (\ (Fct l _ _ ss) -> (l, [])) prg)) })
       state = initState topts defaultRegs prg
       code = do
                --when (Interp.Verbose `elem` topts) $
                --  trace ("++ initState with special registers " ++ (show regs) ++ " and trace opts " ++ (show topts)) (return ())
                when (Interp.Debug `elem` topts) $
                 (showLabTab >> showFunTab >> showParamTab)
                when (Interp.Debug `elem` topts) $
                  trace ("Defs: " ++ (show defs)) (return ())
                let regs = cmdline_regs  `combineRegs` initDefs defaultRegs defs  
                rm <- initRegs regs
                setR rm
                rm <- getR 
                when (Interp.Debug `elem` topts) $
                  trace ("Registers after initialisation: " ++ (showMap rm)) (return ())
                -- trace ("Executing: " ++ (show (CALL (NAME name) ((TEMP thisReg):(Prelude.map (\ n -> CONST n) es))))) (return ())
                let tp = dummyReg
                res <- eval (CALL (NAME name) ((TEMP tp):(Prelude.map (\ n -> CONST n) es)))
                m <- getM
                t <- getT
                c <- getClock
                return (reverse t,
                        c,length (Map.keys m),
                        res)
  in
      if (isNothing codeOk)
        then runIntp state code
        else error ("Errors in checking structure of the code:\n"++(fromJust codeOk))

	                             

dummyVal :: Val
dummyVal = I 99

dummyReg = "t9999"

msg :: String -> Intp ()
msg str = T $ \ (e,m,lT,fT,fsT,fP,t,s,c,p) -> ((e,m,lT,fT,fsT,fP,str:t,s,c,p), ())

msgs :: [String] -> Intp ()
msgs xs = sequence_ (Prelude.map msg xs)

getR = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st,e)
getM = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st,m)
getT = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st,t)
getS = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st,s)
getSpecials = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st,p)
getClock = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st,c)
labTab = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st,lT)
funTab = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st,fT)
fsTab = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st,fsT)
paramTab = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st,fP)
funParams = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st,fP)

setR x = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> ((x,m,lT,fT,fsT,fP,t,s,c,p),())
setM x = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> ((e,x,lT,fT,fsT,fP,t,s,c,p),())
setT x = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> ((e,m,lT,fT,fsT,fP,x,s,c,p),())
setP x = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> ((e,m,lT,fT,fsT,fP,t,s,c,x),())

getReg :: Register -> Intp Val
getReg r = getR >>= \ e -> let x = Map.findWithDefault dummyRegVal r e in return x

getRegSafe :: Register -> Intp Val
getRegSafe a = getReg a >>= \ x -> if x==dummyMemVal 
                                     then error ("PANIC: lookup to uninitialised register " ++ (show a)) 
                                     else return x

setReg :: Register -> Val -> Intp ()
setReg r x = getR >>= \ rm -> setR (Map.union (Map.singleton r x) rm)

getMem :: Address -> Intp Val
getMem a = getM >>= \ m -> let x = Map.findWithDefault dummyMemVal a m in  return x
--getMem a = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> (st, Map.findWithDefault 81 a m)

getMemSafe :: Address -> Intp Val
getMemSafe a = getMem a >>= \ x -> if x==dummyMemVal then error ("PANIC: lookup to uninitialised mem cell " ++ (show a)) else return x

getLabCode :: Label -> Intp [Stm]
getLabCode l = labTab >>= \ lT -> let x = Map.findWithDefault [] l lT in  return x

getFunCode :: String -> Intp [Stm]
getFunCode f = funTab >>= \ fT -> let x = Map.findWithDefault [] f fT in  return x

getFS :: String -> Intp (Maybe Integer)
getFS f = fsTab >>= \ fsT -> let x = Map.findWithDefault Nothing f fsT in  return x

getFunParams :: String -> Intp [Access]
getFunParams f = funParams >>= \ fP -> let x = Map.findWithDefault [] f fP in  return x

getTraceOpts :: Intp [TraceOpts]
getTraceOpts = getS >>= \ (topts, _) -> return topts

getLevel :: Intp Int
getLevel = getS >>= \ (_,lvl) -> return lvl

incLevel :: Intp ()
incLevel = T $ \ st@(e,m,lT,fT,fsT,fP,t,s@(trc,l),c,p) -> ((e,m,lT,fT,fsT,fP,t,(trc,l+1),c,p),())

decLevel :: Intp ()
decLevel = T $ \ st@(e,m,lT,fT,fsT,fP,t,s@(trc,l),c,p) -> if l==0
                                                            then error ("PANIC: trying to return in top-level function")
                                                            else ((e,m,lT,fT,fsT,fP,t,(trc,l-1),c,p),())

tick :: Cost -> Intp ()
tick n = T $ \ st@(e,m,lT,fT,fsT,fP,t,s,c,p) -> ((e,m,lT,fT,fsT,fP,t,s,c+n,p),())

rtError str = error ("** PANIC: "++str)
warn str x = trace ("++ WARNING: "++str) x

type BState = (String, [Stm], Map String [Stm], [(String,[Stm])])
newtype LabTabFiller a = B (BState -> (BState, a))

addBB :: Stm -> LabTabFiller ()
addBB bb = B $ \ (currLab, currBB, m, ss) -> ((currLab, currBB++[bb], m, ss),())

storeCurrBB :: LabTabFiller ()
storeCurrBB = B $ \ (currLab, currBB, m, ss) ->
                       if (currLab/="")
                         then -- trace ("$$ nested label " ++ currLab ++ " added to labTab; code:" ++ (show currBB)) $ 
	                      if (currLab `member` m)
                                then error ("Label "++currLab++" already defined")
	                        else ((currLab,currBB,Map.union (Map.singleton currLab currBB) m, ss),()) 
                         else if (Prelude.null currBB)
                                then ((currLab,currBB,m,ss), ())
                                else {- trace ("PANIC: storeCurrBB nuking currBB: " ++ (show currBB)) -} 
                                     ((currLab,currBB,m,ss), ())

-- like storeCurrBB, but if BB doesn't end with a JUMP or CJUMP, add a JUMP
storeCurrBBWithFinalJump :: Label -> LabTabFiller ()
storeCurrBBWithFinalJump l = getCurrBB >>= \ currBB ->
                    if (Prelude.null currBB) 
                      then addBB (JUMP (NAME l) [l]) >> storeCurrBB
                      else case (Prelude.last currBB) of
                            (JUMP _ _) -> storeCurrBB
                            (CJUMP _ _ _ _ _) -> storeCurrBB
		            _ -> addBB (JUMP (NAME l) [l]) >> storeCurrBB

getCurrBB = B $ \ (currLab, currBB, m, ss) -> ((currLab, currBB, m, ss), currBB)

getCurrLabTab = B $ \ (currLab, currBB, m, ss) -> ((currLab, currBB, m, ss), m)

initCtxt :: String -> LabTabFiller ()
initCtxt l = B $ \ (currLab, currBB, m, ss) -> ((l, [], m, ss), ())


pushCtxt :: LabTabFiller ()
pushCtxt = B $ \ (currLab, currBB, m, ss) -> (("", [], m, (currLab, currBB):ss), ())

popCtxt :: LabTabFiller ()
popCtxt = storeCurrBB >> (B $ \ (_, _, m, (l,bb):ss) -> ((l,bb,m,ss),()))

{-
                 if (currLab/="")
                   then trace ("$$ nested label " ++ currLab ++ " added to labTab; code:" ++ (show currBB)) ((l,bb,Map.union (Map.singleton currLab currBB) m, ss),()) 
                   else ((l,bb,m,ss), ())
-}

runLabTabFiller :: LabTabFiller a -> Map String [Stm]
runLabTabFiller (B f) = let ((_,_,m,_),_) = f ("", [], Map.empty, []) in m

instance Monad LabTabFiller where
 return x = B (\ s -> (s, x))
 (B t) >>= f =  B (\ s -> let { (s1,x1) = t s ; B f' = f x1 } in f' s1)

type  Cost = Integer -- in machine cycles

-----------------------------------------------------------------------------
-- Code Checker monad
-- uses Control.Monad.State

data CState = CCST { labTabCC :: Map String [Stm],    -- labels
                     funTabCC :: Map String [Stm]     -- functions
                   }

-- newtype CodeChecker a = State CState a
type CodeChecker a = State CState a

{-
instance Monad CodeChecker where
 return x = CC (\ s -> (x, s))
 (CC t) >>= f =  CC (\ s -> let { (x1,s1) = t s ; CC f' = f x1 } in f' s1)

instance MonadState CodeState CodeChecker  where
 get = CC $ \ s -> (s,s)
 put s = CC $ \ _ -> ((),s)
-}

-- instance MonadReader CState (CodeChecker a)

-- runTra :: TraState -> TraMonad a -> (a, TraState)
-- runTra traS (T f) = f traS 

-- setLT x = CC $ \ st@(lT,fT) -> ((x,fT), ())
-- getLT   = CC $ \ st@(lT,fT) -> (st, lT)
-- setFT x = CC $ \ st@(lT,fT) -> ((lT, x), ())
-- getFT   = CC $ \ st@(lT,fT) -> (st, fT)

initCState :: CState
initCState = CCST { labTabCC = Map.empty, funTabCC = Map.empty }

bonzo :: (Maybe String) -> (Maybe String) -> (Maybe String)
bonzo Nothing Nothing = Nothing
bonzo Nothing b@(Just s) = b
bonzo b@(Just s) Nothing = b
bonzo (Just s1) (Just s2) = Just (unlines [s1,s2])

runCodeChecker :: CodeChecker (Maybe String) -> (Maybe String)
runCodeChecker f = evalState f initCState

-----------------------------------------------------------------------------
-- Main class, collecting worker functions over the code

class Evaluator x where
 fillTab :: x -> LabTabFiller ()
 cost :: x -> Cost
 eval :: x -> Intp Val
 checkCode :: x -> CodeChecker (Maybe String)

 {-
 prints :: [x] -> IO ()
 prints [] = return ()
 prints (e1:es) = runIntp emptyState (eval e1 >>= \ (_,t1,_) -> eval es >>= \ (_,t2,_) -> return (t1++"\n"++t2++"\n")) >>= \ str -> putStrLn str
 -}

instance Evaluator Stm where
 eval c@(MOVE dest src) = do 
                           -- trace ("Entering " ++ (show c) ++"\n") return ()
                           v <- eval src
                           access <- lval dest
                           store access v
                           topts <- getTraceOpts 
                           let traceme = MoveTrace `elem` topts
                           let warnme = Warnings `elem` topts
                           let rmaps = RegMaps `elem` topts
                           let mmaps = MemMaps `elem` topts
                           when traceme $
                              trace (".. MOVE " ++ (show access) ++ " <- " ++ (show v))  return ()  
                           when rmaps $
                             case (access) of {
                               (InReg r) -> getR >>= \ regs -> trace (".. Registers after " ++ (render (pp c)) ++ "\n" ++ showTree regs ++"\n") (return ()) ;
                                (InMem _) -> return () }
                           when mmaps $
                             case (access) of {
                               (InReg r) -> return () ;
                               (InMem _) -> showAllMem >>= \ str -> trace ("All Mem Dump:\n" ++ str) (return ()) }
                           when warnme $
                            case v of 
                             x@(A a) -> if (x==dummyAddr) then warn ("MOVE: reading uninit Addr " ++ (show x) ++ " from mem " ++ (show access) ++ " in expr " ++ (show src) ) (return ()) else return ()
                             x@(I i) -> if (x==dummyInt) then warn ("MOVE: reading uninit Int " ++ (show x) ++ " from mem " ++ (show access) ++ " in expr " ++ (show src) ) (return ()) else return ()
                           tick (cost c)
                           return dummyVal

 eval c@(EXP e) = tick (cost c) >> eval e 
 eval c@(JUMP dest poss) = do  -- ToDo: test poss
                          (L lab) <- eval dest
                          stm <- getLabCode lab
                          topts <- getTraceOpts 
                          let traceme = JumpTrace `elem` topts
                          let maps = LabMaps `elem` topts
                          if traceme 
                            then trace ("-> JUMP to label " ++ lab ++ (if maps then " continues with code\n" ++ (render $ pp stm) else "")) return ()
                            else return ()
                          tick (cost c)
                          eval stm

 eval c@(CJUMP rel left right trueLab falseLab) =
   do
     x <- eval left
     y <- eval right
     let z = case rel of
              AbsSyn.EQ -> x==y
              AbsSyn.NE -> x/=y
              AbsSyn.LT -> x<y
              AbsSyn.GT -> x>y
              AbsSyn.LE -> x<=y
              AbsSyn.GE -> x>=y
              AbsSyn.ULT -> x<y
              AbsSyn.ULE -> x<=y
              AbsSyn.UGT -> x>y
              AbsSyn.UGE -> x>=y
     
     let (lab, tick_branch) = if z     -- where to continue
                           	 then (trueLab,  tick 1) {- jump only in true case -}
                           	 else (falseLab, tick 0) 
     code <-  getLabCode lab
     topts <- getTraceOpts 
     let traceme = JumpTrace `elem` topts
     let maps = LabMaps `elem` topts
     if traceme 
       then trace ("__ CJUMP results in '" ++ (show x ) ++ " " ++ (show rel) ++ " " ++ (show y) ++ "' which is " ++ (show z)++"\tjumping to label " ++ lab ++ " with code:\n" ++ (render $ pp code)) return ()
       else  return ()
     tick_branch 
     tick (cost code)
     eval code

 -- NB: control-flow handling only works for SEQ-lists, not for arb trees!
 eval c@(SEQ first@(JUMP _ _) second) = do 
                                         tick (cost first)
                             	         eval first
 eval (SEQ first@(CJUMP _ _ _ _ _) second) = do 
                                              tick (cost first)
                             	              eval first
 eval c@(SEQ first second) = do 
                               tick (cost c)
                               eval first
                               eval second

 eval (LABEL l) =  return (L l)

 -- NB: cover expressions, too, since ESEQ might have labels in the first component
 fillTab this@(MOVE dest src) = addBB this >> pushCtxt >> fillTab dest >> popCtxt >>  pushCtxt >> fillTab src >> popCtxt
 fillTab this@(EXP e) = addBB this >>  pushCtxt >> fillTab e >> popCtxt
 fillTab this@(JUMP dest poss) = addBB this >>  pushCtxt >> fillTab dest >> popCtxt
 fillTab this@(CJUMP rel left right trueLab falseLab) = addBB this >>  pushCtxt >> fillTab left >> popCtxt >>  pushCtxt >> fillTab right >> popCtxt
 fillTab this@(SEQ first second) = fillTab first >> fillTab second
 fillTab this@(LABEL l) = do 
                            lT <- getCurrLabTab
                            if (l `member` lT)
                              then error ("Label "++l++" already defined")
                              else storeCurrBBWithFinalJump l >> initCtxt l 

 checkCode (MOVE dest src) = checkCode dest >>= \ ok1 -> checkCode src >>= \ ok2 -> return (bonzo ok1 ok2)
 checkCode (EXP e) = checkCode e
 checkCode (JUMP dest poss) = checkCode dest
 checkCode (CJUMP rel left right trueLab falseLab) = checkCode left >>= \ ok1 -> checkCode right >>= \ ok2 -> return (bonzo ok1 ok2)
 checkCode (SEQ first second) = checkCode first >>= \ ok1 -> checkCode second >>= \ ok2 -> return (bonzo ok1 ok2)
 checkCode (LABEL l) = do
                         s <- get  
                         let lT = labTabCC s 
                         let ok = not ((trc ("%% checking LABEL "++l) l) `member` lT)
                         put $ CCST { labTabCC = Map.union (Map.singleton l []) lT, 
                                      funTabCC = funTabCC s }
                         -- return (trace ("%% LABEL "++l++" OK? \t"++(show ok)) ok)
                         return (if ok 
                                   then Nothing 
                                   else (Just $ "Label "++l++" already defined"))

 -- Cost metric
 cost (MOVE dest src) = 1
 cost (EXP e) = cost e
 cost (JUMP dest poss) = 1
 cost (CJUMP rel left right trueLab falseLab) = cost left + cost right + 1{- rel -} {- + 1 if result is true; assumes fall-through for false -}
 cost (SEQ first second) = cost first + cost second{- unless it's a jump!! -}
 cost (LABEL l) = 0

intOp :: AbsSyn.BinOp -> Integer -> Integer -> Integer
intOp AbsSyn.PLUS x y  = x+y
intOp AbsSyn.MINUS x y  = x-y
intOp AbsSyn.MUL x y  = x*y
intOp AbsSyn.DIV x y  = x `div` y
intOp AbsSyn.AND x y  = boolToVal ((valToBool x)&&(valToBool y))
intOp AbsSyn.OR x y  = boolToVal ((valToBool x)||(valToBool y))
intOp AbsSyn.LSHIFT x y  = x*(mypow 2 y)
intOp AbsSyn.RSHIFT x y  = x `div` (mypow 2 y)
intOp AbsSyn.XOR x y  = boolToVal (myxor (valToBool x) (valToBool y))
intOp op x y = error ((show op) ++" not yet implemented")

instance Evaluator Exp where
 --eval :: Env -> Expr -> (Val, Env)
 eval (CONST value) = return (I value)
 eval (NAME label) = return (L label)
 eval (TEMP temp) = do
                     x <- getRegSafe temp
                     return x
 eval c@(BINOP op left right)=do
                             lhs <- eval left
                             rhs <- eval right
                             tick (cost c)
                             return $
                               case lhs of
                                 (I x) -> case rhs of
                                           (I y) -> (I (intOp op x y))
                                           (A a) -> (A (intOp op x a))
                                 (A a) -> case rhs of
                                           (I y) -> (A (intOp op a y))
                                           (A b) -> (A (intOp op a b))
 eval c@(MEM addr) = do 
                      x <- (eval addr)
                      case x of
                       (A a) -> do
                         z <- getMemSafe a
                         tick (cost c)
                         return z
                       _ -> rtError ("Value of a MEM is not an address:\n  "++(show c)++"; val="++(show x))
                                {-
                            	fpReg' <- fpReg -- name of register
                            	spReg' <- spReg
                            	fp <- getReg fpReg'-- value of register
                            	sp <- getReg spReg'
                                rtError ("Value of a MEM is not an address:\tFP ("++fpReg'++") ="++(show fp)++"\tSP ("++spReg'++") = "++(show sp)++"\n  "++(show c)++"; val="++(show x))
                                -}

 eval c@(CALL func args) = do
                            -- trace ("Entering " ++ (show c) ++ "\n") return ()
                            (L f) <- eval func
                            q <- catchExternal f args
                            case q of 
                              (Just a) -> return a
                              Nothing -> do
                            	fc <- getFunCode f
                            	ps <- getFunParams f
                            	fpReg' <- fpReg
                            	spReg' <- spReg
                            	hpReg' <- hpReg
                            	fp <- getReg fpReg'
                            	sp <- getReg spReg'
                            	hp <- getReg hpReg'
                                old_regs <- getR -- remember old register to simulate callee saving
                                -- old_mem <- getM -- NO: side effects on heap possb
                            	-- trace (".. Allocating frame; current frame at " ++ (show fp) ++"\n") return ()
                                unless ((length ps)==(length args)) $
                                  error ("** call " ++ (show f) ++ " mismatching arg lengths: " ++ (show ps) ++ (show args)) -- (return ())
                            	vals <- passArgs'' ps args -- vals for tracing only
                            	rm <- getR
                                mm <- getM
                                level <- getLevel
                                let str = take level (repeat ' ')
                            	topts <- getTraceOpts 
                            	let traceme = CallTrace `elem` topts
                            	let maps = RegMaps `elem` topts
                            	when traceme $
                            	  trace ("=> " ++ str ++ "CALL   " ++ (show f) ++ (show vals) ++ "\tFP= " ++ (show fp) ++ "\tSP= " ++ (show sp) ++ "\tHP= " ++ (show hp)) return ()
                                         -- (if maps then "\nRegMap after passing args: " ++ (render $ pp args) ++ " -> " ++ (show ps) ++"\n" ++ (showTree rm) else "")++
                                         -- (if maps && any isInFrame ps then "\nMemMap after passing args: " ++ (render $ pp args) ++ " -> " ++ (show ps) ++"\n" ++ (showTree mm) else "")) return ()
                                when (Debug `elem` topts) $
                                  trace ("%% BODY is\n" ++ (render (pp fc))) (return ())
                                incLevel
                                fsOpt <- getFS f -- get frame size     
                            	unless (StdCallConv `elem` topts) $
                                  allocFrame (if isNothing fsOpt then globalFrameSizeHACK else fromJust fsOpt)
                                {-
                            	if traceme 
                            	  then getReg fpReg >>= \ fp -> getReg spReg >>= \ sp -> trace ("** new FP= " ++ (show fp) ++"\tnew SP= " ++ (show sp) ++ "\n") return ()
                                  else return ()
                                -}
                            	eval fc 
                            	rm' <- getR
                                rp' <- retReg
                            	r <- getRegSafe rp' 
                            	unless (StdCallConv `elem` topts) $
                            	  deallocFrame
                                {-
                            	if traceme 
                            	  then getReg fpReg >>= \ fp -> getReg spReg >>= \ sp -> trace ("** old FP= " ++ (show fp) ++"\told SP= " ++ (show sp) ++ "\n") return ()
                                  else return ()
                                -}
                                decLevel
                            	when traceme $
                            	  trace ("=> "++str++"RETURN " ++ (show f) ++ (show vals) ++ " returns value " ++ (show r) ++ (if Debug `elem` topts then " with REG MAP\n" ++ (showTree rm') else "")) (return ())
                                -- setM old_mem
                            	unless (StdCallConv `elem` topts) $
                                  restoreCalleeSavedRegs old_regs
                                tick (cost c)
                            	return r

                         {- let
                           argvals = scanl (foldl (\ (v,q,t) x. eval t q x) (dummyVal,q,t) args
                           (q',t') = bindArgs (thd3 (last argvals)) (snd3 (last argvals)) argvals
                         in
                           eval t' q' (funTab func) -}

 eval c@(ESEQ stm exp) = tick (cost c) >> eval stm >> eval exp

 fillTab (CONST value) = return ()
 fillTab (NAME label) = return ()
 fillTab (TEMP temp) = return ()
 fillTab (BINOP op left right)= pushCtxt >> fillTab left >> popCtxt >> pushCtxt >> fillTab right >> popCtxt
 fillTab (MEM addr) =  return ()
 fillTab (CALL func args) =  pushCtxt >> fillTab args >> popCtxt
 fillTab (ESEQ stm exp) =  pushCtxt >> fillTab stm >> popCtxt >> pushCtxt >> fillTab exp >> popCtxt

 -- code checker
 checkCode (CONST value) = return Nothing
 checkCode (NAME label) = return Nothing
 checkCode (TEMP temp) = return Nothing
 checkCode (BINOP op left right)= checkCode left >>= \ ok1 -> checkCode right >>= \ ok2 -> return (bonzo ok1 ok2)
 checkCode (MEM addr) = checkCode addr
 checkCode (CALL func args) = do
                                ok1 <- checkCode (trc ("%% checking CALL to "++(show func)) func)
                                s <- get
                                let fT = funTabCC s
                                let f = getNameFromExp func
                                let ok2 = if (f `member` fT)
                                  	   then  Nothing
	                          	   else  if f `Prelude.elem` external_fcts 
                                                   then Nothing
						   else Just ("Call to undefined function "++f)  
                                oks <- mapM checkCode args
                                return (let ok = foldl bonzo Nothing ([ok1,ok2]++oks)
                                        in  ok )
 checkCode (ESEQ stm exp) =  checkCode stm >>= \ ok1 -> checkCode exp >>= \ ok2 -> return (bonzo ok1 ok2)

 -- Cost metric
 cost (CONST value) = 0
 cost (NAME label) = 0
 cost (TEMP temp) = 0
 cost (BINOP op left right)= cost left + cost right + 1 {- depending on binop!! -}
 cost (MEM addr) = cost addr
 cost (CALL func args) = sum (Prelude.map cost args) + 3 + (toInteger (length calleeSaves)) {- depends on calling convention! -}
 cost (ESEQ stm exp) =  cost stm + cost exp 


getNameFromExp  (CONST value) = error ("Trying to call a CONST exp")
getNameFromExp  (NAME label) = label
getNameFromExp  (TEMP temp) = error ("Trying to call a TEMP exp")
getNameFromExp  (BINOP op left right)= error ("Trying to call a BINOP exp")
getNameFromExp  (MEM addr) = error ("Trying to call a MEM exp")
getNameFromExp  (CALL func args) = error ("Trying to call a CALL exp")
getNameFromExp  (ESEQ stm exp) =  getNameFromExp exp 

instance Evaluator Fct where
 eval (Fct l ps _ ss) = error "Can't eval function declaration"
 fillTab (Fct l ps _ ss) = fillTab ss
 checkCode (Fct l ps _ ss) = do
                             -- let lT = runLabTabFiller (fillTab ss >> getCurrLabTab)
                             s <- get
                             let fT = funTabCC s
                             -- put $ CCST { labTabCC = lT , funTabCC = fT }
                             -- HACK: empty list encodes an existing function that hasn't been met by checkCode, yet
                             let b1 = if ((trc ("%% checking function "++l) l) `member` fT && (not (Prelude.null (fT!l)))) 
                               		then Just ("Function "++l++" already defined") 
                               		else Nothing
                             put $ CCST { labTabCC = labTabCC s , 
                                          funTabCC = (Map.union (Map.singleton l ss) fT)  }
                             b2 <- (checkCode ss)
                             return (bonzo b1 b2) 
 cost _ = 0

instance Evaluator [Stm] where
 eval [] =  getTraceOpts  >>= \ topts -> when (Debug `elem` topts) (trace "%% done (Stm)" (return ())) >>  return  dummyVal
 eval (x@(JUMP _ _):xs) = eval x
 eval (x@(CJUMP _ _ _ _ _):xs) = eval x
 eval (x:xs) = eval x >>  getTraceOpts  >>= \ topts -> when (Debug `elem` topts) (trace ("%% CODE now\n" ++ (render (pp x)) ++ "\nrest of CODE [Stm]\n" ++ (render (pp xs))) (return ())) >> eval xs
 fillTab [] = storeCurrBB >> initCtxt ""
 fillTab (x:xs) = fillTab x >> fillTab xs
 -- checkCode [] = return True
 checkCode xs = mapM checkCode xs >>= \ oks -> return (foldl bonzo Nothing oks)
 cost = sum . (Prelude.map cost)

instance  Evaluator [Exp] where
 eval [] =  getTraceOpts  >>= \ topts -> when (Debug `elem` topts) (trace "%% done (Stm)" (return ())) >> return dummyVal
 -- DEBUG: eval (x@(CALL _ _):xs) = trace ("%% ABOUT to CALL " ++ (render (pp x)) ++ "\nCODE now\n" ++ (render (pp x)) ++ "\nrest of CODE [Exp]\n" ++ (render (pp xs))) (return ()) >> eval x >> trace ("%% CODE now\n" ++ (render (pp x)) ++ "\nrest of CODE [Stm]\n" ++ (render (pp xs))) (return ()) >> eval xs
 eval (x:xs) = eval x >> getTraceOpts  >>= \ topts -> when (Debug `elem` topts) (trace ("%% CODE now\n" ++ (render (pp x)) ++ "\nrest of CODE [Stm]\n" ++ (render (pp xs))) (return ())) >> eval xs
 fillTab [] = storeCurrBB >> initCtxt ""
 fillTab (x:xs) = fillTab x >> fillTab xs
 -- checkCode [] = return True
 checkCode xs = mapM checkCode xs >>= \ oks -> return (foldl bonzo Nothing oks)
 cost = sum . (Prelude.map cost)
 
---------------------------------------------------------------------------
-- Mem mgmt:

data Access = InReg Register
            | InFrame (Integer, Integer) -- (Frame, Offset)
            | InMem Address
 deriving (Read,Show,Eq)

isInRegister (InReg _) = True
isInRegister    _ = False
isInFrame  (InFrame _) = True
isInFrame    _ = False
isInMem    (InMem _) = True
isInMem    _ = False

lval :: Exp -> Intp Access
lval (TEMP r) = return (InReg r)
lval (MEM e) = do
                a <- eval e
                case a of { (A a) -> return (InMem a)  
                          ; (L l) -> rtError ("lval returned label " ++ l) 
                          ; (I l) -> rtError ("lval returned integer " ++ (show l)) }
lval e@(ESEQ stm exp) = 
              do
                topts <- getTraceOpts 
                when (Interp.Verbose `elem` topts) $
                     trace (".. Found complex expression as lval: "++(show e)) (return ())
                eval stm
                lval exp
lval e = error ("Illegal lval: '"++(render $ pp e)++"'")

store :: Access -> Val -> Intp Val
store (InReg r) x = do 
                      rm <- getR 
                      setR (Map.union (Map.singleton r x) rm) 
                      return (dummyVal)
store (InFrame (0,o)) x = do
                            mm <- getM
                            fp <- fpReg
                            qq <- getReg fp 
                            case qq of { (A fpa) -> setM (Map.union (Map.singleton (fpa+o) x) mm) 
                                       ; (I _) -> rtError ("store: Trying to store InFrame to Int rather than Addr") }
                            return (dummyVal)
store (InFrame (1,o)) x = do 
                            mm <- getM
                            sp <- spReg 
                            qq <- getReg sp
                            case qq of { (A spa) -> setM (Map.union (Map.singleton (let a'=(spa+o) in {-trace ("__ FrameStore: "++(show x)++" -> "++(show (A a')))-} a') x) mm) 
                                       ; (I _) -> rtError ("store: Trying to store InFrame to Int rather than Addr") } 
                            return (dummyVal)
store (InMem a) x = do
                      mm <- getM
                      fp <- fpReg
                      f <- getReg fp
                      setM (Map.union (Map.singleton a x) mm)
                      return (dummyVal)
store a x = rtError $  "store: illegal access of destination: "++(show a)

storeA (A a) x = T $ \ (e,m,lT,fT,fsT,fP,t,s,c,p) -> ((e,Map.insert a x m,lT,fT,fsT,fP,t,s,c,p), dummyVal)
storeA _ x = error "Cannot store into a non-address"

param2access :: AbsSyn.Param -> Access
param2access (LOC n) = InFrame (1,n) -- 1 means one frame off; i.e. take sp as basis in store!!
param2access (REG t) = InReg t
--param2access (ADR a) = InMem t

showMem :: Address -> Address -> Intp String
showMem from to | to <= from = return ""
showMem from to | otherwise = 
  do
    m <- getMem from
    str <- showMem (from+4{-wordSize-}) to
    return ((show from) ++ " -> " ++ (show m) ++ "\n" ++ str)

showAllMem :: Intp String
showAllMem = hpReg >>= \ hpReg' -> getReg hpReg' >>= \ x -> getSpecials >>= \ (_,_,_,_,_,hsz) ->
  let hpSt = if (hsz==0) 
               then case hpEnd of (A s) -> (s-defHpSz)
               else case hpEnd of (A s) -> (s-hsz)
  in
   case x of 
    (A xa) -> showMem hpSt xa
    (I i) -> rtError ("showAllMem: heap pointer register contains Int rather than Addr: " ++ (show (I i)))

-- NB: this is simultaneous assignment, otw old vars might be overwritten
passArgs :: [Access] -> [Exp] -> Intp [Val]
passArgs [] [] = return []
passArgs (p:ps) (e:es) = do 
	                   trace ("++ passArgs " ++ (show e) ++ " -> " ++ (show p)) (return ())
	                   showRegMap
	                   x <- eval e
                           store p x
	                   xs <- passArgs ps es
                           topts <- getTraceOpts 
                           let traceme = ParamTrace `elem` topts
                           when traceme $
                             trace ("passArgs'': " ++ (show e) ++ (show x) ++ " -> " ++ (show p)) (return ())
                           return (x:xs)

-- works only for registers now!!
passArgs' :: [Access] -> [Exp] -> Intp [Val]
-- passArgs' [] [] = return []
passArgs' ps es = do
			   vals <- mapM eval es
	                   rm <- getR
                           let alist = Prelude.zip (Prelude.map ( \ (InReg r) -> r) ps) vals
	                   setR (Map.union (Map.fromList alist) rm) -- works only for InRegs
                           topts <- getTraceOpts 
                           let traceme = ParamTrace `elem` topts
                           when traceme $
                             trace ("passArgs': " ++ (Prelude.concat (Prelude.map (\ (x,y) -> (show y) ++ " -> " ++ (show x) ++ "  ") alist))) (return ())
	                   return vals

passArgs'' :: [Access] -> [Exp] -> Intp [Val]
-- passArgs' [] [] = return []
passArgs'' ps es =do
		   vals <- mapM eval es
		   rm <- getR
	           sp <- spReg
	           fp <- fpReg
              	   f@(A fpa) <- getReg fp
              	   s@(A spa) <- getReg sp
        	   let alist = Prelude.zip ps vals
		   mapM ( \ (p,v) -> store p v )  alist
        	   -- store setR (Map.union (Map.fromList alist) rm) -- works only for InRegs
        	   topts <- getTraceOpts 
        	   let traceme = ParamTrace `elem` topts
        	   when traceme $
        	     trace ("passArgs''"++(if any isInFrame ps then "\tFP = "++(show f)++"\t"++" SP = "++(show s)++"\n"  else "\n")++
                            (unlines (Prelude.map (\ (x,y) -> (show y) ++ " -> " ++ (show x) ++ "  ") alist))) (return ())
		   return vals

-- passArgs' _ _ = rtError "passArgs': Mismatch of lengths of ars and param lists"

allocFrame :: Integer ->  Intp ()
allocFrame sz = do
	         sp <- spReg
	         fp <- fpReg
              	 s@(A spa) <- getReg sp
              	 f@(A fpa) <- getReg fp
	         when (s+(I sz)<spEnd) $
                   rtError "stack overflow"
                 store (InMem spa) f -- (A fpa)
	      	 setReg fp (A spa)
	      	 setReg sp (A (spa-sz))

deallocFrame :: Intp ()
deallocFrame = do
	        sp <- spReg
	        fp <- fpReg
              	(A spa) <- getReg sp
                (A fpa) <- getReg fp
	        setReg sp (A fpa)
	        m <- getMem fpa
	        setReg fp m

initRegs :: SpecialRegs -> Intp (Map String Val)
initRegs p@(fp,sp,hp,rp,ssz,hsz) = 
 do 
  let spSt = if (ssz==0) 
               then spStart
               else case spEnd of (A s) -> A (s+ssz) -- -(spStart-spEnd)
  let hpSt = if (hsz==0) 
               then case hpEnd of (A s) -> A (s-defHpSz)
               else case hpEnd of (A s) -> A (s-hsz)
  let xs = Prelude.map ( \ x -> (x, dummyVal) ) calleeSaves 
  --   rp <- retReg
  --   fp <- fpReg
  --   sp <- spReg
  --   hp <- hpReg
  topts <- getTraceOpts 
  when (Interp.Verbose `elem` topts) $
    trace ("Chosen Registers: FP="++fp++"; SP="++sp++"; HP="++hp++"; RP="++rp++"; SSZ="++(show ssz)++"; HSZ="++(show hsz)++"\nSTACK: ("++(show spSt)++") - ("++(show spEnd)++")\tHEAP: ("++(show hpSt)++") - ("++(show hpEnd)++")") (return ())
  setP p
  return $ Map.fromList ([(rp, dummyVal), (fp, spSt), (sp, spSt), (hp, hpSt), (dummyReg, dummyVal)]++xs)
   -- return (foldr Map.union Map.empty xs') 

initDefs :: SpecialRegs -> [(String, Either String Integer)] -> SpecialRegs
initDefs regs xs = 
  foldl (\ r@(fp,sp,hp,rp,sz,hz) (s,v') -> 
           case v' of {
             -- SET NAME: put new name in specials
             Left v -> case trc ("SETting NAME of "++s++" to "++v) s of { "fp" -> (v ,sp,hp,rp,0,0)
                                                                  ; "sp" -> (fp,v ,hp,rp,sz,hz)
                                                                  ; "hp" -> (fp,sp, v,rp,sz,hz)
                                                                  ; "rp" -> (fp,sp,hp, v,sz,hz)
                                                                  --; "sz" -> (fp,sp,hp,rp,(read v),hz)
                                                                  --; "hz" -> (fp,sp,hp,rp,      sz,(read v))
                                                                  ; _ -> error ("PANIC: initDefs: unknown abstract machine setting for "++s++" in SET NAME "++s++" = "++v) }
            -- SET VALUE: should be used in initRegs 
            ; Right n -> case trc ("SETting VALUE of "++s++" to "++(show n)) s of { "sz" -> (fp,sp,hp,rp, n,hz)
                                                                    ; "hz" -> (fp,sp,hp,rp,sz, n)
                                                                    ; _ -> trace ("WARNING: ignoring SET VALUE for "++s) r } } )
  regs xs

-- all regs EXCEPT hp are callee saved
restoreCalleeSavedRegs :: RegMap -> Intp ()
restoreCalleeSavedRegs old_regs = 
  do 
    hpReg' <- hpReg
    hp <- getReg hpReg'
    setR old_regs
    setReg hpReg' hp
    return ()
		        
---------------------------------------------------------------------------
-- handling of external calls

external_fcts = ["L_halloc_obj", "L_init_obj", "L_halloc_arr", "L_init_arr", "L_print", "L_raise"]

catchExternal :: String -> [Exp] -> Intp (Maybe Val)
catchExternal "L_halloc_obj" es = do
                                  vals <- mapM eval es 
                                  a <- case vals of
                                    	 [(I n)] -> halloc_obj n
                                    	 _ -> rtError ("illegal arguments to _halloc_obj (should be 1 integer argument): " ++ (show vals))
                                  return (Just a)
catchExternal "L_init_obj" es = do
                                  vals <- mapM eval es 
                                  case vals of
                                    [(A a), (I n)] -> init_obj a n
                                    _ -> rtError ("illegal arguments to _init_obj (should be 1 address and 1 integer argument): " ++ (show vals))
                                  return (Just dummyAddr)
catchExternal "L_halloc_arr" es = do
                                  vals <- mapM eval es 
                                  a <- case vals of
                                    	 [(I n)] -> halloc_arr n
                                    	 _ -> rtError ("illegal arguments to _halloc_arr (should be 1 integer argument): " ++ (show vals))
                                  return (Just a)
catchExternal "L_init_arr" es = do
                                  vals <- mapM eval es 
                                  case vals of
                                    [(A a), (I n)] -> init_arr a n
                                    _ -> rtError ("illegal arguments to _init_arr (should be 1 address and 1 integer argument): " ++ (show vals))
                                  return (Just dummyAddr)
catchExternal "L_print" es = do
                                  vals <- mapM eval es 
	                          msgs (Prelude.map show vals)
                                  return (Just dummyAddr)
catchExternal "L_raise" es = do
                               vals <- mapM eval es 
                               case vals of
                                   [(I n)] -> error ("Runtime error: "++(show n))
                               return (Just dummyAddr)
catchExternal _ es = return Nothing -- error ("Unknown external function "++f)

-- allocate an object of given size
halloc_obj :: Integer -> Intp Val
halloc_obj sz = do
	          hpp <- hpReg
	          hp@(A hpa) <- getReg hpp
                  topts <- getTraceOpts 
                  let traceme = AllocTrace `elem` topts
                  when traceme $
	            trace (".. Allocating object at address " ++ (show (A hpa)) ++ "\tsize: " ++ (show sz)) (return ())
	          let new_hp = (A (hpa+sz+4)) -- Andreas, 2007-11-08 +4 for ClassID
	          when (new_hp>hpEnd) $
                   rtError "heap overflow"
	          hpReg' <- hpReg
	          setReg hpReg' new_hp
	          return hp

init_obj :: Address -> Integer -> Intp ()
init_obj a sz = do
	              --(A a) <- obj
	              mm <- getM
	              setM (Map.union (Map.fromList (Prelude.map (\ x -> (a + 4 * x,(A 0))) [0..(sz `div` 4)])) mm)
	              return ()

-- allocate an array of given length
halloc_arr :: Integer -> Intp Val
halloc_arr len = do
	          hpp <- hpReg
	          hp@(A hpa) <- getReg hpp
                  topts <- getTraceOpts 
                  let traceme = AllocTrace `elem` topts
                  when traceme $
	            trace (".. Allocating array at address " ++ (show (A hpa)) ++ "\tlength: " ++ (show len)) (return ())
	          let new_hp = (A (hpa+4*(1+len))) -- Andreas, 2007-11-08 +4 for length
	          when (new_hp>hpEnd) $
                    rtError "heap overflow"
	          hpReg' <- hpReg
	          setReg hpReg' new_hp
	          return (hp+(I 4))  -- NB: length at index: -1

init_arr :: Address -> Integer -> Intp ()
init_arr a len = do
	              --(A a) <- obj
	              mm <- getM
                      setM (Map.union (Map.fromList ((a-4,(I len)):(Prelude.map (\ x -> (a + 4 * x, (I 0))) [0..len-1]))) mm)  -- Andreas, 2007-11-07, 2008-02-13 initialize with zeros
-- BUGGY ORIGINAL:
--	              setM (Map.union (Map.fromList ((a-1,(I sz)):(Prelude.map (\ x -> (x,(I 99))) [a..a+sz]))) mm)
	              return ()

-- DUMMY
addrOfFrame _ = 0

mypow m 0 = 1
mypow m n = m*(mypow m (n-1))

myxor x y = x/=y

valToBool 0 = False
valToBool _ = True

boolToVal False = 0
boolToVal True = 1
