-- Time-stamp: <Tue Jan 08 2008 16:34:04 Stardate: [-29]8783.24 hwloidl>
-- Interpreter for the intermediate language used in 
-- Appel's "Modern Compiler Implementation in Java"
---------------------------------------------------------------------------

module Main where

import System
import System.Console.GetOpt
import Control.Monad
import Text.Regex
import List

import AbsSyn
import Lex
import Parse
import PPIntm
import Interp

data Flag = Verbose
          | ShowAST 
          | Help 
          | Traces String
          | CallT
          | JumpT
          | MoveT
          | AllocT
          | ParT
          | RegM
          | MemM
          | LabM
          | Debug
          | FP String
          | SP String
          | SSZ String
          | HSZ String
          | HP String
          | RP String
          | StdCallConv
          | Warnings
          | Output String
            deriving (Eq, Show)

optDescrs :: [OptDescr Flag]
optDescrs =
  [ Option ['?'] ["help"]      (NoArg Help)     "show usage information"
  , Option ['v'] ["verbose"]   (NoArg Main.Verbose) "be verbose"
  , Option ['a'] ["ast"]       (NoArg Main.ShowAST) "show AST"
  , Option ['T'] ["defTrace"]  (ReqArg Traces "Traces") "default tracing flags (NOT IMPLEMENTED)"
  , Option ['C'] ["callTrace"] (NoArg CallT) "trace function calls"
  , Option ['J'] ["jumpTrace"] (NoArg JumpT) "trace jumps"
  , Option ['M'] ["moveTrace"] (NoArg MoveT) "trace moves"
  , Option ['A'] ["allocTrace"](NoArg AllocT) "trace memory allocations"
  , Option ['P'] ["paramTrace"](NoArg ParT) "trace parameter passing"
  , Option ['R'] ["regMap"]    (NoArg RegM) "print register maps"
  , Option ['Q'] ["memMap"]    (NoArg MemM) "print memory maps"
  , Option ['L'] ["labMap"]    (NoArg LabM) "print label maps"
  , Option ['D'] ["debugging"] (NoArg Main.Debug) "debug interpreter itself"
  , Option ['f'] ["framepointer"]  (ReqArg FP "fp") "name of frame pointer register"
  , Option ['s'] ["stackpointer"]  (ReqArg SP "sp") "name of stack pointer register"
  , Option ['h'] ["heappointer"]   (ReqArg HP "hp") "name of heap pointer register"
  , Option ['r'] ["returnvalue"]   (ReqArg RP "rp") "name of this pointer register"
  , Option ['S'] ["stacksize"]     (ReqArg SSZ "size") "stack size"
  , Option ['H'] ["heapsize"]      (ReqArg HSZ "size") "heap size"
  , Option ['W'] ["warnings"]      (NoArg Main.Warnings) "print warnings"
  , Option ['z'] ["stdcallconv"]   (NoArg Main.StdCallConv) "standard calling convention (default: all regs are callee saved)"
  , Option ['o'] ["output"]    (ReqArg Output "FILE") "output file (NOT IMPLEMENTED)"
  ]

usage :: [String] -> IO a 
usage errs = ioError (userError (concat errs ++ usageInfo header optDescrs ++ "$Revision: 1.11 $")) 
    where header = "Usage: muHwI [options] <file>" 

{-
data Options = Options {
    verbose :: Bool,
    traces :: [String]
  }

emptyOpts :: Options
emptyOpts = Options False []
-}

setByFlag :: Flag -> TraceOpts
setByFlag CallT = CallTrace
setByFlag JumpT = JumpTrace
setByFlag MoveT = MoveTrace
setByFlag AllocT = AllocTrace
setByFlag ParT = ParamTrace
setByFlag RegM = RegMaps
setByFlag MemM = MemMaps
setByFlag LabM = LabMaps
setByFlag Main.ShowAST = Interp.ShowAST
setByFlag Main.Verbose = Interp.Verbose
setByFlag Main.Debug = Interp.Debug
setByFlag Main.Warnings = Interp.Warnings
setByFlag Main.StdCallConv = Interp.StdCallConv
setByFlag _ = None

setRegisters :: Flag -> SpecialRegs
setRegisters (FP fp) = (fp,"","","",0,0)
setRegisters (SP sp) = ("",sp,"","",0,0)
setRegisters (HP hp) = ("","",hp,"",0,0)
setRegisters (RP rp) = ("","","",rp,0,0)
setRegisters (SSZ sz) = ("","","","",read sz,0)
setRegisters (HSZ sz) = ("","","","",0,read sz)
setRegisters _ = ("","","","",0,0)

parseCmdLine :: [String] -> IO (String,[TraceOpts],SpecialRegs,[String])
parseCmdLine argv = do
  let (os, ns, errs) = getOpt Permute optDescrs argv
  if (length ns)<1
    then usage [""]
    else return ()
  let (prgFile:args) = ns
  when (Main.Verbose `elem` os) $ putStrLn $ "Reading program from file: " ++ prgFile
  when (Main.Help `elem` os) $ ioError (userError "No execution with help")
  return (prgFile, 
          filter (/=None) (map setByFlag os), 
	  foldl combineRegs undefinedRegs  (map setRegisters os),  
          args)

main =	do
  cmdLine  <- getArgs
  (prgFile, topts, regs, args) <- parseCmdLine cmdLine
  input <- readFile prgFile
  -- putStrLn $ "Passing these args to main fct: " ++ (show args)
  -- input <- getContents
  -- print (alexScanTokens input)
  let (defs,ast) = parse (alexScanTokens input)
  when (Interp.ShowAST `elem` topts) $
   putStrLn ("AST:\n" ++ (render $ pp ast))
  --let res = Interp.runIntp (Interp.initState ast) (Interp.eval 
  let rx = mkRegex "^.*\\$"
  -- let xs = map (\ (AbsSyn.Fct name _ _) -> (subRegex rx name "")) ast
  -- putStrLn (unlines xs)
  let main_ast = case (find (\ (AbsSyn.Fct name _ _ _) -> (subRegex rx name "") == "main") ast) of 
   		   Nothing -> error "PANIC: no main function found" 
   		   (Just main_ast) -> main_ast 
  let (msgs,c,m,res) = Interp.execCall_with_env topts defs regs ast main_ast (map read args)
  putStrLn ("Messages:\n" ++ unlines msgs)
  -- putStrLn ("Result:\n" ++ (show res))
  -- NB: computing cycles creates huge stack sizes; turned off for now
  -- putStrLn ("Execution time: " ++ (show c) ++ " cycles;\t Heap size: " ++ (show m) ++ " cells")
